import {inject, ref, Ref, watch} from "vue"
import {LocationQueryValue, RouteLocationNormalizedLoaded, Router} from "vue-router"

export type TabEntry = {
  id: string,
  scrollDestId?: string
}

export type AppendTabFunction = (entry: TabEntry) => void
export type RemoveTabFunction = (entry: TabEntry) => void
export type SelectTabEventHandler = (event: { id: string, scrollDestId?: string }) => void

export type TabHandler = {
  currentTabIndex: Ref<number>
  currentTabId: Ref<string | undefined>
  appendTab: AppendTabFunction
  removeTab: RemoveTabFunction
  selectTab: SelectTabEventHandler
}

export type ScrollToTabOption = {
  offset: number
}

export const TAB_HANDLER = Symbol("TabHandler")

export const useTabHandler = (): TabHandler | undefined => inject(TAB_HANDLER)

/**
 * `UiTabs` 를 router 에 연결해 사용할 때 필요한
 * `selectedTabId` 변수와 `changeRouterTab` 함수를 리턴한다.
 *
 * ```html
 * <UiTabs v-model="selectedTabId" @update:modelValue="changeRouterTab">
 * ```
 *
 * ```ts
 * setup() {
 *   return {
 *     ...useRouterTab(router, route, 'tab')
 *   }
 * }
 * ```
 * @param router Router 객체. `useRouter()` 로 얻을 수 있다.
 * @param route Route 객체. `useRoute()` 로 얻을 수 있다.
 * @param defaultTabId router query 에 탭이 없을 때 디폴트 값으로 사용되는 tab id.
 * @param paramName router query 에 추가될 파라미터 명
 */
export const useRouterTab = (router: Router, route: RouteLocationNormalizedLoaded, defaultTabId: string, paramName = 'tab'): {
  selectedTabId: Ref<string | LocationQueryValue[] | null>
  changeRouterTab: (tabId: string, tabs: Ref<TabEntry[]>) => void
} => {

  const selectedTabId = ref(route.query[paramName])
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  const changeRouterTab = (tabId: string, tabs: Ref<TabEntry[]>): void => {
    const query = {...route.query}
    // 디폴트 탭이 선택된 경우는 query param 을 넣지 않는다. (처음 진입 시 query 가 변경되는 문제를 방지하기 위함)
    if (tabId === defaultTabId) {
      delete query[paramName]
    } else {
      query[paramName] = tabId
    }
    router.push({path: route.path, query})
  }
  watch(() => route.fullPath,
      () => selectedTabId.value = route.query[paramName] || defaultTabId)

  return {
    selectedTabId,
    changeRouterTab
  }
}
